(* Spheres.ml: this is free software by Christophe Raffalli *)
(* ocamlopt -I +lablGL lablgl.cmxa lablglut.cmxa str.cmxa unix.cmxa spheres.ml -o spheres *)
(* see: http://paulbourke.net/miscellaneous/stereographics/stereorender/ *)

let version = "1.1"
let compat = "1.1" (* oldest readable version *)

open Printf
open Scanf
open Objects_luke_skywalker

(* This program either write a game file, and read one to replay an existing game *)
type ('a,'b) file =
    Out of 'a
  | In of 'b

(* A type to store the initial parameters of the program *)
type options = {
  stereo : bool;
  file : (string, string) file;
  seed : int; pseudo : string;
  reg : bool;
  init_width : int; init_height : int;
  details : int;
  mutable refresh : int; (* number of refresh per second *)
}

(* Pass options to Glut and then read the remaining options ...
   Note: the window dimension in -geometry are ignored *)
let params = Glut.init Sys.argv

let options_spec o = [
  ("-3d", Arg.Unit (fun () -> o := { !o with stereo = true}), " Activates anaglygh 3D for cyan/red glasses.");
  ("-out", Arg.String (fun s -> o := { !o with file = Out s}),
   " Specifies the file name for saving the game (default: \".spheres\").");
  ("-in", Arg.String (fun s -> o := { !o with file = In s}),
   " Gives the file name storing a game to be replayed.");
  ("-p", Arg.String (fun s -> o := { !o with pseudo = s}),
   " Specifies your pseudo for saving the game to the website (default: value of the USER environment variable");
  ("-s", Arg.Int (fun i -> o := { !o with seed = i}),
   " Gives the seed of the random number genrator. The same seed alwayas yeld the same game (default: 0).");
  ("-w", Arg.Int (fun i -> o := { !o with init_width = i}), " Gives the initial screen width (default: 800).");
  ("-h", Arg.Int (fun i -> o := { !o with init_height = i}), " Gives the initial screen height (default: 800).");
  ("-d", Arg.Int (fun i -> o := { !o with details = i}), " Gives the details level (default: 30).");
  ("-r", Arg.Int (fun i ->
                    if i < 10 then raise (Arg.Bad "Refresh must me at least 10");
                    if i > 100 then raise (Arg.Bad "Refresh greater than 100 is useless !");
                    o := { !o with refresh = i}), " Gives the number or refresh per second (default: 35).");
(* no HTTP in this version
  ("-db", Arg.Unit (fun () -> o := { !o with reg = true}),
   " Register game to the website (default: off).");
*)
  ]

let options =
  let opts = ref {
    stereo = false; file = Out ".spheres"; seed = 0;
    pseudo = Sys.getenv "USER"; reg = false; refresh = 35;
    init_width = 800; init_height = 800; details = 30;
  } in
  try
    Arg.parse_argv params (Arg.align (options_spec opts))
      (fun s -> raise (Arg.Bad ("Don't know what to do with: \""^s^"\"")))
      "usage: params.(0) options";
    !opts
  with
    Arg.Help s -> prerr_endline s; exit 0
  | Arg.Bad s ->  prerr_endline s; exit 1

(* Open the file either to save the game or to read a game *)
let file =
  let name = match options.file with Out f | In f -> f in
  try match options.file with
    Out file ->
      Random.init options.seed;
      let ch = open_out file in
      fprintf ch "%dR%d" options.seed options.refresh;
      Out (ch, ref None)
  | In file ->
      let ch = Scanning.from_file file in
      bscanf ch "%dR%d"
        (fun seed r ->
           options.refresh <- r;
           Random.init seed);
      In (ch, ref None)
  with
    Sys_error s ->
      fprintf stderr "Can not open file \"%s\": %s.\n" name s;
      flush stderr;
      exit 1
  | Scan_failure s ->
      fprintf stderr "Invalid game file \"%s\" (or version older than %s): %s.\n" name compat s;
      flush stderr;
      exit 1

(* a filter function that either write current speed to ch if file = Out ch or
   read the speed from ch if file = In ch and discard its input *)
let read_write_move xs ys =
  let float_to_char = function
    | 0.0 -> 'O' | 1.0 -> 'A' | -1.0 -> 'M' | _ -> assert false
  in
  let char_to_float = function
    | 'O' -> 0.0 | 'A' -> 1.0 | 'M' -> -1.0 | _ -> assert false
  in
  match file with
  | Out(_, ({ contents = None } as ptr)) ->
      ptr := Some(1,xs,ys);
      (xs, ys)
  | Out(_, ({ contents = Some(n,xs',ys') } as ptr)) when xs = xs' && ys = ys' ->
      ptr := Some(n+1,xs,ys);
      (xs, ys)
  | Out(ch, ({ contents = Some(n,xs',ys') } as ptr)) ->
      fprintf ch "%c%c%d" (float_to_char xs') (float_to_char ys') n;
      ptr := Some(1,xs,ys);
      (xs, ys)
  | In(ch, ({contents = (None | Some(0, _, _))} as ptr)) ->
      let xs, ys, n =
        try
          bscanf ch "%c%c%d" (fun xc yc n ->
                                let xs = char_to_float xc and ys = char_to_float yc in
                                xs, ys, n)
        with
          Scan_failure _ ->
            prerr_endline "Invalid game of mistamatch in OCaml version";
            exit 1
      in
      ptr := Some(n-1,xs,ys);
      xs, ys
  | In(ch, ({contents = Some(n, xs, ys)} as ptr)) ->
      ptr := Some(n-1,xs,ys);
      xs, ys

(* End the game
   1) finalise and save the game file, or when reading, check that the score agree.
   2) if -db : post the game on the website
*)
let end_of_game ship =
  Printf.printf "points = %d\n" ship.our_points;
  begin
    match file with
      Out(ch,moves) ->
        (* force writing last moves and register score *)
        begin
          match !moves with
          | Some(n,xs,ys) when n > 0 ->
              ignore
                (read_write_move
                   (if xs = 0.0 then 1.0 else 0.0) (if ys = 0.0 then 1.0 else 0.0))
          | _ -> ()
        end;
        fprintf ch "S%d\n" ship.our_points;
        close_out ch;
    | In(ch,_) ->
        (* check score *)
        try
          bscanf ch "S%d" (fun n -> if n <> ship.our_points then raise Exit);
        with
          Scan_failure _ | Exit->
            prerr_endline "Invalid game of mistamatch in OCaml version";
            exit 1
  end;
  (* no HTTP in this version
  if options.reg then begin
    let file = match options.file with
        In file | Out file -> file
    in
    try
      let ch = open_in file in
      let game = input_line ch in
      close_in ch;
      match
        Http.post "pseudo=%s&score=%d&seed=%d&game=%s&version=%s"
          options.pseudo ship.our_points options.seed game version
      with
        Http.Accepted "OK" ->
          prerr_endline "Game written to data base."
      | Http.Accepted str ->
          prerr_endline str
      | Http.Failed str ->
          prerr_endline (sprintf "Failed to write game to data base (%s)." str)
    with _ ->
      prerr_endline "Failed to read game file."
  end;
  *)
  exit 0

(* We store the current width and height in global references and create
   the callback that informs openGL of resizing *)
let cur_width = ref options.init_width
let cur_height = ref options.init_height

let reshape ~w ~h =
  cur_width := w;
  cur_height := h;
  GlDraw.viewport ~x:0 ~y:0 ~w ~h;
  Gl.flush ()

(* Data and type for camera *)
type camera =
  { aperture : float; near : float;
     far : float; focallength : float;
     mutable eyesep : float }

let camera =
  { aperture = 45.0;  near = 0.75;
     far = 30.0; focallength = 1.0;
     eyesep = (if options.stereo then 0.05 else 0.0) }

(* frustum declaration is missing in the glut.mli file !!! *)
external frustum :
    x:(float * float) -> y:(float * float) -> z:(float * float) -> unit
    = "ml_glFrustum"

(* to support 3D anaglyph, we create one functions setting up the camera allowing
   a shift. If should be called with (0.5 *. camera.eyesep) for the left eye and
   the opposite for the right one.
   FIXME : corner should be corrected
   FIXME : 3D anaglyph needs testing *)
let set_camera  =
  let pi = atan 1.0 *. 4.0 in
  let radians = pi /. 180.0 *. (camera.aperture /. 2.0) in
  let wd2 = camera.near *. tan(radians) in (* this two lines are not so clear ...
                                              why is there a relation between near plane and
                                              projection matrice *)
  let ndfl = camera.near /. camera.focallength in
  let top = wd2 and bottom = -. wd2 in
  (fun shift ->
    GlMat.mode `projection;
    GlMat.load_identity ();
    let ratio = float !cur_width /. float !cur_height in
    let left  = -. ratio *. wd2 +. shift *. ndfl in
    let right  = ratio *. wd2 +. shift *. ndfl in
    frustum ~x:(left,right) ~y:(bottom,top) ~z:(camera.near,camera.far);
    GluMat.look_at
    ~eye:(-. shift, 0.0, 1.0)
    ~center:(-. shift, 0.0, 0.0)
    ~up:(0.0, 1.0, 0.0);      (* up is in positive Y direction *)
    GlMat.mode `modelview;
    GlMat.load_identity ();
  )

(* display text information and the cross mark in the middle of the creen *)
let display_info s x_pos y_pos =
  GlLight.material ~face:`front (`emission (1.0,1.0,1.0,1.0));
  let cross_half_size = 0.04 in
  GlPix.raster_pos ~x:cross_half_size ~y:cross_half_size ~z:0.0 ();
  String.iter (fun c ->
    Glut.bitmapCharacter Glut.BITMAP_HELVETICA_12 (Char.code c)) s;
  GlDraw.begins `lines;
  GlDraw.vertex3 (-.cross_half_size+.x_pos*.cross_half_size, 0.0, 0.0);
  GlDraw.vertex3 (cross_half_size+.x_pos*.cross_half_size, 0.0, 0.0);
  GlDraw.vertex3 (0.0, -.cross_half_size+.y_pos*.cross_half_size, 0.0);
  GlDraw.vertex3 (0.0, cross_half_size+.y_pos*.cross_half_size, 0.0);
  GlDraw.ends();
  GlLight.material ~face:`front (`emission (0.0,0.0,0.0,0.0))

(* Need to create window before GL-compiling the initial sphere, so do know
   most of OpenGl initialization (including lighting) *)
let _ =
  Glut.initDisplayMode ~depth:true ~double_buffer:true ~alpha:false ();
  Glut.initWindowSize !cur_width !cur_height;
  ignore(Glut.createWindow("spheres"));
  GlLight.light ~num:0  (`diffuse (1.0, 0.0, 0.0, 1.0));
  GlLight.light ~num:0  (`ambient (1.0, 1.0, 1.0, 0.));
  GlLight.light ~num:0  (`specular (1.0, 1.0, 1.0, 0.));
  GlLight.light ~num:0  (`position (1.0, 1.0, 2.0, 0.));
  Gl.enable `lighting;
  Gl.enable `light0;
  GlLight.material ~face:`front (`ambient (0.1,0.1,0.1,1.0));
  GlLight.material ~face:`front (`shininess 0.5);
  Gl.enable `depth_test

(* This function populates the glList for faster rendering *)
let compile_sphere s =
  let qobj =
    GluQuadric.create () in
  GluQuadric.draw_style qobj `fill ;
  GluQuadric.normals qobj `smooth;
  GlLight.material ~face:`front (`diffuse s.sphere_color);
  GlLight.material ~face:`front (`specular s.sphere_color);
  GlMat.rotate3 77.3 (1. , 1. , 1.2);
(*  Glut.solidTeapot s.sphere_radius;*)
  let tesselation = max 3 (truncate (float options.details *. sqrt s.sphere_radius)) in
  GluQuadric.sphere ~quad:qobj ~radius:s.sphere_radius
    ~slices:tesselation  ~stacks:tesselation ();
  GlList.ends ();
  Gc.finalise (fun s -> GlList.delete s.sphere_user_data) s

let spheres =
  let s = first_sphere (GlList.create `compile) in
  compile_sphere s;
  ref [s]

(* reference to compute fps every 5s *)
let last_fps = ref 0.0
let period_start_fps = ref (Unix.gettimeofday ())
let refresh_nb = ref 0

let get_update_fps () =
  let cur_time = Unix.gettimeofday () in
  let diff = cur_time -. !period_start_fps in
  if diff >= 5.0 then begin
    last_fps := float !refresh_nb /. diff;
    refresh_nb := 1;
    period_start_fps := cur_time;
  end else incr refresh_nb;
  !last_fps

(* function to display all spheres *)
let display_spheres () =
  List.iter (fun sphere ->
               GlMat.push ();
               GlMat.translate3 (Vector.sub sphere.sphere_pos ship.our_pos);
               GlList.call sphere.sphere_user_data;
               GlMat.pop ();
            ) !spheres

(* global display function *)
let displayFunc () =
  (* compute the string of information to display *)
  let (x,y,_) = ship.our_pos in
  let str = string_of_int ship.our_points in
  (* clear the screen *)
  GlFunc.color_mask ~red:true ~green:true ~blue:true ();
  GlClear.clear [`color; `depth];
  (* left eye or global display *)
  set_camera (0.5 *. camera.eyesep);
  if options.stereo then
    GlFunc.color_mask ~red:true ~green:false ~blue:false ();
  display_spheres ();
  display_info str x y;
  (* right eye display if stereo *)
  if options.stereo then begin
    set_camera (-0.5 *. camera.eyesep);
    GlClear.clear [`depth];
    GlFunc.color_mask ~red:false ~green:true ~blue:true ();
    display_spheres ();
    display_info str x y;
  end;
  Glut.swapBuffers()

(* references for accurate timing *)
let start_time = ref 0.0
let num_step = ref 0.0
let refresh = float options.refresh
let alarm_step = 1000.0 /. refresh
let pause = ref false

let rec timefunc ~value =
  (* restart the timer, compensating for errors *)
  let time = Unix.gettimeofday () in
  let n =  !num_step in
  if n <= 0.0 then start_time := Unix.gettimeofday ();
  let error = n *. alarm_step -. 1000.0 *. (time -. !start_time) in
  if error <= -1000.0 then begin
    prerr_endline "Computer too slow sorry.";
    exit 1;
  end;
  let step = max 10 (truncate (alarm_step +. error)) in
  Glut.timerFunc step timefunc 1;
  num_step := n +. 1.0 ;

  if not !pause then begin
    (* update ship position *)
    let (x,y,z) = ship.our_pos in
    let clamp x = if x < -1.0 then -1.0 else if x > 1.0 then 1.0 else x in
    let (xs, ys) = read_write_move ship.our_x_speed ship.our_y_speed in
    ship.our_pos <- (
      clamp (x +. xs /. refresh), clamp (y +. ys /. refresh), z);

    (* create a new_sphere with expected proba *)
    let new_sphere =
      if Random.float 1.0 < level.new_proba /. refresh then
        let s = random_sphere (GlList.create `compile) in
        compile_sphere s;
        avoid_others s !spheres;
        [s]
      else
        []
    in

    let remaining_spheres =
      try
        move_spheres_and_check refresh ship !spheres;
      with
        End_of_game -> end_of_game ship
    in

    spheres := new_sphere @ remaining_spheres;
    check_collisions_between !spheres;
    increase_level refresh;
  end;
  Glut.postRedisplay()

(* Call back for keyboard: splitted in two in GLUT *)
let specialKey down ~key ~x ~y =
  match key, down with
  | Glut.KEY_LEFT, true -> ship.our_x_speed <- -1.0
  | Glut.KEY_LEFT, false -> ship.our_x_speed <- 0.0
  | Glut.KEY_RIGHT, true -> ship.our_x_speed <- 1.0
  | Glut.KEY_RIGHT, false -> ship.our_x_speed <- 0.0
  | Glut.KEY_UP, true -> ship.our_y_speed <- 1.0
  | Glut.KEY_UP, false -> ship.our_y_speed <- 0.0
  | Glut.KEY_DOWN, true -> ship.our_y_speed <- -1.0
  | Glut.KEY_DOWN, false -> ship.our_y_speed <- 0.0
  | Glut.KEY_END, _ -> exit 0;
  | _ -> ()

(* Function to change to fullscreen.Remember the
   screen position in a reference.
   if b is false only exit from fullscreen mode
   otherwise change more *)
let is_fullScreen = ref None

let fullScreen b =
  match !is_fullScreen, b with
    None, true ->
      is_fullScreen := Some(!cur_width,!cur_height,
                              Glut.get Glut.WINDOW_X,
                              Glut.get Glut.WINDOW_Y);
      Glut.positionWindow 0 0;
      Glut.fullScreen ()
  | Some(w,h,x,y), _ ->
      is_fullScreen := None;
      Glut.reshapeWindow w h;
      Glut.positionWindow x y
  | _ -> ()

let keyboardFunc ~key ~x ~y =
  let c = Char.chr key in
  match c with
  | 'q' | 'Q' -> exit 0
  | 'p' | 'P' -> pause := not !pause
  | 'f' | 'F' -> fullScreen true
  | '\027' (* escape *) ->
      fullScreen false
  | 'e' ->
      let n = camera.eyesep *. 1.05 in
      camera.eyesep <- n;
      Printf.printf "eyesep = %f\n" n; print_newline ();
  | 'E' ->
      let n = camera.eyesep /. 1.05 in
      camera.eyesep <- n;
      Printf.printf "eyesep = %f\n" n; print_newline ();
  | c ->
      Printf.fprintf stderr "invalid key: %c.\n" c; flush stderr

(* register all callbacks and start main loop *)
let _ =
  Glut.reshapeFunc reshape;
  Glut.displayFunc displayFunc;
  Glut.specialUpFunc (specialKey false);
  Glut.specialFunc (specialKey true);
  Glut.keyboardFunc keyboardFunc;
  Glut.keyboardUpFunc (fun ~key ~x ~y -> ());
  Glut.timerFunc 100 timefunc 1;
  Glut.mainLoop()
